// Advanced_faceStep.cpp: implementation of the CFaceStep class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "scribble.h"
#include "ScribVw.h"
#include "Advanced_faceStep.h"
#include "Face_boundStep.h"
#include "RepresentationStep.h"
#include "stepser.h"
#include "Edge_loopStep.h"// wh 10/17/01
#include "SurfaceStep.h"

#include "Cartesian_pointStep.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CAdvanced_faceStep ::CAdvanced_faceStep( CBaseStep*pIn,
                                         CFace_boundStep* p_face_bound, 
                                         bool same_sense
                                       , int iNumFB
                                       )
: CFace_surfaceStep( pIn, p_face_bound, iNumFB)
{
  nameClass("Advanced_face");
  m_bSame_sense = same_sense;
  m_pFace_geometry = dynamic_cast< CSurfaceStep*> (pIn);
  p_face_bound->m_pBack = this;
}

CAdvanced_faceStep::~CAdvanced_faceStep ()
{

}

HRESULT CAdvanced_faceStep::backPtr(CBaseStep *&prBrep)
{
  if (NULL!=m_pBack){
    m_pBack->backPtr( prBrep );  
    return S_OK;
  }
  else {
    return E_FAIL;
  }
}

void CAdvanced_faceStep::Serialize(stepArchive &ar)
{
  try{
   if (!isSerialized()){
      setSerialized();
      writePre(ar);
      ar <<   "'', " ;
      writeNtuple(ar, &m_bounds, m_bounds.GetSize());
      ar <<        ",#" ;
      ar << m_pFace_geometry->getNum();// face outer bound
      ar <<   ", " ;
      ar << m_bSame_sense ;
      writePost(ar);

#ifdef  WRITEEULER
      CBaseStep* pBaseStep=NULL;
      backPtr( pBaseStep);
      if (NULL!=pBaseStep){
        CManifold_solid_brepStep* pManifold_solid_brepStep=(CManifold_solid_brepStep*) pBaseStep;
        pManifold_solid_brepStep->eulerCnt.m_numFaces++;
      }
#endif//WRITEEULER

      for (int i = 0; i < m_bounds.GetSize(); i++){
        CFace_boundStep* pFB = (CFace_boundStep*) m_bounds.GetAt(i);
        ASSERT( this == pFB->m_pBack);
        pFB->Serialize(ar);
      }
      m_pFace_geometry->m_pBack = this;
      m_pFace_geometry->Serialize(ar);
    }
  }
  catch(...){
    ar << "\n/*Errors detected in CAdvanced_faceStep ::Serialize #\n";
//  ar << getNum();
    ar <<   ", but continuing*/\n" ;    
  }
}

HRESULT CAdvanced_faceStep::AddFB( CFace_boundStep *p_face_bound
                              , int &ri
                              )
{
  HRESULT hr = S_OK;
#if 0
  // before adding a ring, we must 
  //  first check if this ring is contained by another ring. 
  //  second, we must check if it contains another ring

  CPlaneStep* pPlaneStep = dynamic_cast<CPlaneStep*>
    (this->m_pFace_geometry);

  if (NULL != pPlaneStep ){
    CDirectionStep* pFaceLeft = pPlaneStep->m_pPosition->m_pRef_direction;

    for (int i = 1; i < m_bounds.GetSize(); i++){
      CFace_boundStep* pFBThis = (CFace_boundStep*) m_bounds.GetAt(i);
      if (pFBThis->m_bOrientation && pFBThis->m_bEnabled){
        CEdge_loopStep * pPLStepIn = dynamic_cast<CEdge_loopStep *>(p_face_bound->m_bound);
      
        boundsIntersectBounds rBoundsXbounds = unknownBoundsIntersection;
        CEdge_loopStep * pPLStepThis = dynamic_cast<CEdge_loopStep *>(pFBThis->m_bound);

        unsigned char bitContainmentO = unknownContainment;
        pPLStepThis->contains                                         
          ( (pPLStepIn->GetEdge_list())
          , bitContainmentO 
          , pFaceLeft
          );
        if (in == bitContainmentO ) {
          m_bounds.InsertAt( i, (CObject*)p_face_bound);
          CFace_boundStep *p_fb = (CFace_boundStep *)m_bounds.GetAt( i + 1);
//          p_fb->m_bEnabled = false;
          ri = i;
          return hr;
        }

        unsigned char bitContainmentI = unknownContainment;
        pPLStepIn->contains
          ( (pPLStepThis->GetEdge_list())
          , bitContainmentI
          , pFaceLeft
          );
        if (in == bitContainmentI) {// to really be safe
          // don't add it because it is in a hole unless there is a surrounding ring
          p_face_bound->m_bEnabled = false;
          m_bounds.InsertAt( i, (CObject*)p_face_bound);
          ri = -1;
          return hr;
        }
      }// fi m_bOrientation
    }// next i
  }
  ri = m_bounds.Add((CObject*)p_face_bound);
#endif
  return hr;
}

HRESULT CAdvanced_faceStep::trimEdge( CrcPtrArray *pIntersectionPts  // edge w/ matter on LHS
                                   , paramPtrArray& intersectionEdge  // those points of pIntersectionPts which are within bounds
                                   , containment& bContains  // is this pt contained
                                   )
{
  HRESULT hr = S_OK;
//  ASSERT(pIntersectionPts->GetSize() > 1);
  // make sure that the intersection is more than just a nick
  if (pIntersectionPts->GetSize() > 1){

    // hold onto this because list of face bounds will grow!
    int numFBs = this->m_bounds.GetSize();
//  numFBs = 1;// hard set until the code is all working
    for (int i = 0; i < numFBs; i++)
    {
      // the first FB is an FOB
      CFace_boundStep* pFB = ((CFace_boundStep*) this->m_bounds.GetAt(i));
      CFace_outer_boundStep* pFOB = dynamic_cast<CFace_outer_boundStep*> (pFB );
      if (NULL == pFOB ) {
        continue;
      }
      // create a new fob which will be the other side of the split
      CFace_outer_boundStep*  pFOB2 = 
        new CFace_outer_boundStep( NULL, 
                                   NULL,
                                   true // wh 02/07/02
                               //, NULL
                                 );
//    pFOB2->m_bEnabled = false;
      pFOB2->m_bound = new CEdge_loopStep();// wh 11/05/01

      // now we split the original fob
      hr = pFOB->trimEdge( pIntersectionPts
                         , intersectionEdge
                         , ((CPlaneStep*)m_pFace_geometry)->m_pPosition->m_pAxis 
                         , bContains
                         );

      if (NULL == pFOB2 ){
        hr = E_OUTOFMEMORY;// not really needed since CMemoryException will be thrown
      }
      else {
        if ( 0 < ((CEdge_loopStep*)(pFOB2->m_bound))->GetEdge_list().GetSize()){
          // add the new fob to the list of face bounds for this surface
          this->m_bounds.Add( pFOB2);// this->m_bounds.InsertAt( i++, pFOB2);
          pFOB2->demote(); // wh 02/07/02
          pFB->demote(); // wh 02/07/02
          ASSERT(pFB->m_bEnabled == true);// wh 02/10/02
        } else {
          // degenenate case
          // we really don't need it, so get rid of it
          pFOB2->release();
        }
      }// fi FOB
  
    }// next i
  }// fi
  return hr;
}
// determine if the point is contained within the bounds
HRESULT CAdvanced_faceStep::contains( CCartesian_pointStep*  pPoint
                                    , containment&           bContains
                                    , double&                rdParam
                                    )
{
#if 1
  CDirectionStep* pFaceNormal = NULL;
  this->m_pFace_geometry->normal( NULL, pFaceNormal );
  this->GetFace_outer_bound()->contains( pPoint, bContains, pFaceNormal
     , rdParam
    );
#endif
  return S_OK;
}

// calculate the face-edge intersection
// each edge is oriended ccw, henc wh just have to make sure that they are properly inserted into the list
HRESULT CAdvanced_faceStep::edgeXAdvanced( CEdge_loopStep*    pELoopN,      // loop corresponding to neg feature
                                           double&            dStartParam,  // start parameter on line0
                                           double&            dEndParam,    // end parameter on line0
                                           CrcPtrArray*       pLoopIntersectionPoints,  // collection of intersection 
                                                                                      //points on polygon
                                           unsigned int &     riStartIndex, // index of startpoint
                                           unsigned int &     riEndIndex    // index of endpoint
                                         , CCartesian_pointStepPtrArray*& pCartesian_pointStepPtrArray
                                         )
{ 
  HRESULT hr = S_OK;

  int iInsertionIndex = 0;
  int iAdditionIndex = 0;// how far from the end do you insert these??

  for (int iv = 0; iv < pELoopN->GetSize(); iv++){
    // create an edge on neg features bounded , that edge is oriented ccw, 
    //  hence the intersections are also oriented ccw as well, 
    //  provided that the last one was not encountered first!
    CPlaneStep* pPlaneStepOnPosFeature = dynamic_cast<CPlaneStep*>
      (this->m_pFace_geometry);
    if (NULL != pPlaneStepOnPosFeature ) {
//    pPlaneStepOnPosFeature ->edgeXAdvanced();
   // int iNumIntersection=0;
    // create an edge on neg features bounded 

    COriented_edgeStep* pOriented_edgeStep = pELoopN->GetAt(iv  );

    // intersect this edge with the plane
    if( (NULL != pOriented_edgeStep)  
      ){

      CVertex_pointStep* pVertex_pointStep0 = dynamic_cast<CVertex_pointStep*>
        (pOriented_edgeStep->GetEdge_start()) ;
      CCartesian_pointStep* pPointStep0 = dynamic_cast<CCartesian_pointStep*>
        (pVertex_pointStep0->m_p_vertex_geometry);
      
      CVertex_pointStep* pVertex_pointStep1 = dynamic_cast<CVertex_pointStep*>
        (pOriented_edgeStep->GetEdge_end()) ;
      CCartesian_pointStep* pPointStep1 = dynamic_cast<CCartesian_pointStep*>
        (pVertex_pointStep1->m_p_vertex_geometry);

      intersectionType intersection = noIntersection;//unknownSegmentIntersection 
      double paramU, paramV,
             paramT[2];

      CrcPtrArray* pEdgeIntersectionPoints = NULL;
      try{
        pEdgeIntersectionPoints = new CrcPtrArray();
      } catch (...){
        return E_OUTOFMEMORY;
      }
    
      CEdge_curveStep *pEdge_curveStep = dynamic_cast<CEdge_curveStep*>
        (pOriented_edgeStep->m_pEdge_element);
      CLineStep *pLineStep = dynamic_cast<CLineStep*>
        (&(pEdge_curveStep->m_rEdge_geometry));

      hr = 
        pPlaneStepOnPosFeature->intersect( pLineStep , 
                                           intersection, 
                                           pEdgeIntersectionPoints, 
                                           paramU, 
                                           paramV,
                                           paramT );
      if (FAILED( hr)){

//      delete pEdge;

        return hr;
      }
      if ( noIntersection != intersection){
/*
  //    CDirectionStep faceLeft(NULL, pEdge->m_dMagnitude );
        containment            bContains=unknownContainment;
        double dParam;
        // wait, not so fast, we have to make sure that intersection point is inside FOB
        pELoopN->contains( (CCartesian_pointStep*)pEdgeIntersectionPoints->GetAt(0), bContains, (pEdge->m_pOrientation), dParam);
        if (out==bContains){
          intersection = noIntersection ;
        }
*/
#if 1
        if ( startpointIntersection == intersection) {
          // find out where the surface is to the left of the traversal direction
          // if edge vector is antiparellel to surface normal, start point
          // if edge vector is parallel to surface Normal, endpoint
          double dp = 
            dotProduct( *(pPlaneStepOnPosFeature->m_pPosition->m_pAxis)
            , *(pLineStep->m_pDir->m_pOrientation)
            );
          if (dp>0){
//          if (pLoopIntersectionPoints->GetAt( pLoopIntersectionPoints->GetSize()-1) != (void*)pCartesian_pointStep0){
              riStartIndex = pLoopIntersectionPoints->Add((void*)pLineStep->m_pPos);
//          }
//          else {
//            riStartIndex = pLoopIntersectionPoints->GetSize()-1;
//          }
          } else if (dp<0){
            riStartIndex = 0;
//          if (pLoopIntersectionPoints->GetAt( iInsertionIndex) != (void*)pCartesian_pointStep0){
              pLoopIntersectionPoints->InsertAt( iInsertionIndex, (void*)pLineStep->m_pPos);
              iInsertionIndex++;          
//          }
          } else {
            TRACE0("coincidence??\n");
          }
        }  
        else if ( endpointIntersection == intersection) {
          // find out where the surface is to the left of the traversal direction
          // if edge vector is antiparellel to surface normal, start point
          // if edge vector is parallel to surface Normal, endpoint
          double dp = 
            dotProduct( *(pPlaneStepOnPosFeature->m_pPosition->m_pAxis)
                      , *(pLineStep->m_pDir->m_pOrientation)
                      );
          if (dp>0){
            riEndIndex = pLoopIntersectionPoints->Add((void*)pLineStep->m_pPos1);
          } else if (dp<0){
            riEndIndex = 0;
            pLoopIntersectionPoints->InsertAt(iInsertionIndex , (void*)pLineStep->m_pPos1);
            iInsertionIndex++;
          } else{
            TRACE0("coincidence??\n");
          }
        }  
        else if ( interiorIntersection == intersection) {

          // small problem in that intersection is sometimes unnecssary 
          //  because the loop goes outside the outer_bound and 
          //  has not been contained as in MERGE_BOUNDS_INTERSECTION_ADV ! // wh 02/25/02

          // accept only intersections between the start_point and end_point.of Edge1.
          // Not only do we have to find the params, 
          //   but we also have to split the loop on face of the neg feature and truncate it
          CCartesian_pointStep *pCartesian_pointOld = (CCartesian_pointStep *)pEdgeIntersectionPoints->GetAt(0);
          CCartesian_pointStep *pCartesian_pointNew = pCartesian_pointOld;
  /*
          // need to search against the list of known points, and if found use the other one!  //wh 11/13/01   
          pCartesian_pointStepPtrArray->Find( pCartesian_pointOld, pCartesian_pointNew);// wh 11/13/01
          if (NULL != pCartesian_pointOld){
            pCartesian_pointStepPtrArray->Add( pCartesian_pointNew);// wh 11/13/01
          }
  */
          // we need to make sure the intersection points are ordered CCW
          // if edge vector is antiparellel to surface normal, start point
          // if edge vector is parallel to surface Normal, endpoint
          double dp = 
            dotProduct( *(pPlaneStepOnPosFeature->m_pPosition->m_pAxis)
                      , *(pLineStep->m_pDir->m_pOrientation)
                      );
          if (dp>restol2){
            pLoopIntersectionPoints->Add((void*) pCartesian_pointNew);
          } else if (dp<-restol2){
            pLoopIntersectionPoints->InsertAt( iInsertionIndex 
              , (void*)pCartesian_pointNew);
            iInsertionIndex++;
          } else {
            TRACE0("coincidence??\n");
            ASSERT(0);
          }

          // we should also add this intersection to the posFeature's face?

  // wh 10/22/01 true

          // wh 10/15/01  setting true gives better results, but gaps_in loops
          // modify the negative feature by adding the vertex at the intersection
  // seem to miss one due to tolerance issues!// wh 11/06/01
            // check if the point is within the [outer]bound
            CDirectionStep* pFaceNormal = pPlaneStepOnPosFeature->m_pPosition->m_pAxis;
            containment bContains = unknownContainment; 
            double rdParam;
            this->contains( 
                  (CCartesian_pointStep*) pEdgeIntersectionPoints->GetAt(0)
                , bContains
                //, pFaceNormal 
                , rdParam                    
                ); 

            if ( out != bContains )// really should be    (on == bContains)
            {
#if 0
              if (0 != iv){ 
                if ((pELoopN->GetAt(iv  ) != pCartesian_pointNew ) && 
                    (pELoopN->GetAt(ivm1) != pCartesian_pointNew ) ){
                  // also double check that this isn't an endpoint

                  pELoopN->InsertAt( iv,(void*)pCartesian_pointNew );
                  iv++;
                }
              }
              else {
                if ((pELoopN->GetAt(iv  ) != pCartesian_pointNew ) && 
                    (pELoopN->GetAt(ivm1) != pCartesian_pointNew ) ){
                  // do as above

                  pELoopN->Add( (void*)pCartesian_pointNew );
                }
              }
              // also need to add them to the pos surface's loops
#endif
            }// fi contain
          // fi MOD_NEG_FEATURE

          // can there be more intersections here??, 
            //probably not on a plane, what about with numerical noise?
        }// fi interiorIntersection == intersection
        else if ( coincidentIntersection == intersection) {
          // should not reveal more info than we already have, but could be used to simplify
          dStartParam = min( dStartParam, paramT[0]);
          dEndParam   = max( dEndParam  , paramT[1]);

// need this to prevent gaps
// wh 3/13/02
/*
          if (-1) {
            faceXedgeAdvancedCoincident( pPlaneStepOnPosFeature
                                      , pELoopN                     // loop corresponding to neg feature
                                      , pLoopIntersectionPoints     // collection of intersection 
                                                                    //     points on polygon
                                      , riStartIndex                // index of startpoint
                                      , riEndIndex                  // index of endpoint
                                      , pCartesian_pointStep0 
                                      , pCartesian_pointStep1
                                      , iInsertionIndex
                                      );

          }
*/
        }
        else {
          for (int i = 0; i< pEdgeIntersectionPoints->GetSize();i++){
             delete pEdgeIntersectionPoints->GetAt(i);
          }
        }
#endif
      }
      else {
      }
      delete pEdgeIntersectionPoints;

//    delete pEdgeN;
//      delete pEdge;

    }
    // if you modified the neg feature, you should split the face into two as well
//    ivm1 = iv;
    } else {
//    CConicStep* pConicStepOnPosFeature = dynamic_cast<CConicStep*>
//      (this->m_pFace_geometry);
//    conicXedgeAdvanced();
    }
  }// next iv

  return S_OK;
}


HRESULT CAdvanced_faceStep::intersect( CAdvanced_faceStep*  pAdvanced_faceStepOnNegFeature
                                     , t_aEdge_loopStep&    aEdge_loopStep                          // returned list of edges on intersection
//                                   , CAdvanced_faceStep*  pAdvanced_faceStepOnPosFeatureModified //post intersection 
//                                   , CAdvanced_faceStep*  pAdvanced_faceStepOnNegFeatureModified //post intersection 
                                     )
{

  CPlaneStep* pPlaneStepOnPosFeature = dynamic_cast<CPlaneStep*>
    (this->m_pFace_geometry);

  if (NULL != pPlaneStepOnPosFeature ){
  
    CPlaneStep* pPlaneStepOnNegFeature  = dynamic_cast<CPlaneStep*>
      (pAdvanced_faceStepOnNegFeature->m_pFace_geometry);
    if (NULL != pPlaneStepOnNegFeature ){

      CDirectionStep* pIntersection= NULL; 
      // calculate the direction of the intersection edge
      pIntersection=crossProduct( *pPlaneStepOnPosFeature->m_pPosition->m_pAxis 
                                , *pPlaneStepOnNegFeature->m_pPosition->m_pAxis
                                );
      bool bPlanesIntersect = !pIntersection->withinTol();  
      if (!bPlanesIntersect ) { 
        if (
             pPlaneStepOnPosFeature->onSurface( 
               *(pPlaneStepOnNegFeature->m_pPosition->m_pLocation))
           ){
          bPlanesIntersect = true;
        }
      }

      // make sure that the faces are not parallel
      if (bPlanesIntersect ){
      
        double dStartParam =  DBL_MAX;//???
        double dEndParam   = -DBL_MAX;//???

        CEdge_loopStep *pEdge_loopStep = new 
          CEdge_loopStep();

//      pEdge_loopStep->AddPoint();
      }

      delete pIntersection;
    }//fi
  }//fi

  return S_OK;
}

//////////////////////////////////////////////////////////
// who  date      history
// wh   12/23/01  Advanced_face moved to its own file
// wh   12/23/01  should the method Intersect modify the class or just calculate the intersection?

